package org.jabber.applet.tools;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.*;

/***************************************************************************
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in
 * either source code or executable form, except in compliance with the 
 * License.  You may obtain a copy of the License at 
 * 
 * http://www.jabber.com/license/ 
 * 
 * or at
 * 
 * http://www.opensource.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the
 * License for the specific language governing rights and limitations 
 * under the License.
 * 
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber and to Jack Harich for his article 
 * in javaworld.com on how to create a reusable MessageBox class.
 ***************************************************************************/

/**
 * This is a generic dialog that can be used in two ways...
 * 1. InputBox: a window that presents a series of 
 *      inputs to the user. 
 * OR
 *
 * 2. Msgbox: a window with a question and one or more buttons
 *    representing the answer.
 *
 * Since a thread is used to set the dialog to visible,
 * when the client calls ask() it will not block.
 *
 * The client may implement ActionListener, which has:
 *      public void actionPerformed(ActionEvent evt)
 * if user response notification is desired.
 *
 * @author Jonathan Paulson
 * @version 0.9.1
 */
public class InputDialog implements Runnable, ActionListener, WindowListener, KeyListener, FocusListener {
    private ActionListener        moListener;
    private Object                moCallback;
    private Dialog                moDialog;
    private String                msCloseWindowCommand    = "CloseRequested";
    private Frame                moFrame;
    private boolean                mbframeNotProvided;
    private Panel                moButtonPanel            = new Panel();
    private Panel                moInputPanel            = new Panel();
    private Panel                moListPanel                = null;
    private Color                moColor;
    private Font                moFont;
    private Font                boldFont;
    private String                msTitle;
    private int                    miQNbr                    = 0;
    private int                    mX                        = -1;
    private int                    mY                        = -1;
    private Vector                mvQuestions                = null;
    private Vector                mvButtons                = null;
    private Vector                mvListValues            = null;
    private String                msButtonWithFocus        = "";
    private String                msDefaultButton            = "";

    /**
     * This constructor is used to delare the
     * listener that will be notified when a button is clicked.
     * The listener must implement ActionListener.
     *
      * @param    parent    parent frame that implements ActionListener
     * @param    color    Dialog color
     * @param    font    Dialog font
     * @param    title    Dialog title
     */
    public InputDialog(Object parent, Color color, Font font, String title) {
        this.moCallback = parent;
        if( parent instanceof Frame ) {
            setFrame( (Frame)parent );
        }
        if( parent instanceof ActionListener ) {
            setActionListener( (ActionListener)parent );
        }
        this.moFont = font;
        // create a bold font for use in captions.
        this.boldFont = new Font( moFont.getName(),
                                  Font.BOLD,
                                  moFont.getSize() );
        this.moColor = color;
        this.msTitle = title;

        GridLayout gridLayout = new GridLayout(1,0);
        moInputPanel.setLayout(gridLayout);
        moInputPanel.setBackground(moColor);

        //Add extra blank label to the left and right of the buttons
        //row so they are not squished against the sides
        //(2nd label added in displayDialog)
        gridLayout = new GridLayout(2,2,5,0);        //was (1,2) todo
        moButtonPanel.setLayout(gridLayout);
        moButtonPanel.setBackground(moColor);
        Label lblSpacer = new Label();
        lblSpacer.setBackground(moColor);
        moButtonPanel.add(lblSpacer);

        mvQuestions = new Vector(20, 10);
        mvButtons = new Vector(10, 10);
        mvListValues = new Vector(20, 10);
    }

    //---------- Runnable Implementation ---------------------
    /**
     * This prevents the caller from blocking, which
     * if this class is used on an awt event thread would
     * cause a deadlock.
     */
    public void run() {
        moDialog.setVisible(true);
    }


    //---------- FocusListener Implementation ---------------
    public void focusGained(FocusEvent e) {

        //Only buttons have focus listeners.

        //Save the text of the button with the focus

        msButtonWithFocus = ((Button)e.getComponent()).getLabel();

    }



    public void focusLost(FocusEvent e) {

        msButtonWithFocus = "";

    }

    //---------- ActionListener Implementation ---------------
    public void actionPerformed(ActionEvent evt) {
        if (moListener != null) {
            //pass on event to Listener
            moListener.actionPerformed(evt);
        }
        closeDialogAndCallBack(evt.getActionCommand());
    }

    //---------- KeyListener Implementation ------------------
    public void keyTyped(KeyEvent evt) {
        Object source = evt.getSource();
        if( source instanceof TextField ) {
            TextField tf = (TextField)source;
            String restricted = tf.getName();
            if( restricted == null )
                return;

            String typed = String.valueOf(evt.getKeyChar());
            if( restricted.indexOf(typed) > -1 ) {
                // java2 handles this correctly - e.g. the
                // keyTyped is called BEFORE the textfield gets
                // the event. Java 1.1 (at least < 1.1.8) does NOT.
                // So, if the user typed a restricted character, it is
                // already in the textfield by this time (Usefull, hunh).
                String tfText = tf.getText();
                if( tfText != null && tfText.endsWith(typed) ) {
                    tfText = tfText.substring(0, tfText.length()-1);
                    tf.setText(tfText);
                    tf.setCaretPosition( tf.getText().length() );
                }
                evt.consume();
                InputDialog iDialog = new InputDialog( null, moColor,
                                                       moFont, "ERROR");
                iDialog.setLocation(mX,mY);
                iDialog.msgboxOK("Cannot use the characters "
                                 +  restricted +
                                 " in this field.");

            }
        }
    }

    public void keyPressed(KeyEvent evt) {
        //close dialog when excape key hit
        if (evt.getKeyCode() == KeyEvent.VK_ESCAPE) {
            fireCloseRequested();
        } else if(evt.getKeyCode() == KeyEvent.VK_ENTER) {
            //Let the [Enter Key] act as a button click
            if (msButtonWithFocus.length() > 0) {
                //close the dialog and act as if button with focus was clicked
                closeDialogAndCallBack(msButtonWithFocus);
            } else if (msDefaultButton.length() > 0) {
                //close the dialog and act as if default button with was clicked
                closeDialogAndCallBack(msDefaultButton);
            }
        }
    }

    public void keyReleased(KeyEvent evt) { }

    private void fireCloseRequested() {
        ActionEvent event = new ActionEvent(this,
                                            ActionEvent.ACTION_PERFORMED, msCloseWindowCommand);
        actionPerformed(event);
    }

    //---------- WindowListener Implementatons ---------------
    public void windowClosing(WindowEvent evt) {
        // User clicked on X or chose Close selection
        fireCloseRequested();
    }

    public void windowClosed(WindowEvent evt) { }
    public void windowDeiconified(WindowEvent evt) { }
    public void windowIconified(WindowEvent evt) { }
    public void windowOpened(WindowEvent evt) { }
    public void windowActivated(WindowEvent evt) {  }
    public void windowDeactivated(WindowEvent evt) { }



    //---------- Public Methods ------------------------------
    /**
     * This sets the listener to be notified of button clicks
     * and WindowClosing events.
     *
     * @param    listener    listener to be notified 
     */
    public void setActionListener(ActionListener listener) {
        this.moListener = listener;
    }

    /**
     * This sets this dialog's title
     *
     * @param    title    Dialog title
     */
    public void setTitle(String title) {
        this.msTitle = title;
    }

    /**
     * If a Frame is provided then it is used to instantiate
     * the modal Dialog. Otherwise a temporary Frame is used.
     * Providing a Frame will have the effect of putting the
     * focus back on that Frame when the MessageBox is closed
     * or a button is clicked.
     *
     * @param frame        Frame to be used to focus back on when 
     *                    frame closed
     */
    public void setFrame(Frame frame) { // Optional
        this.moFrame = frame;
    }

    /**
     * Sets the ActionCommand used in the ActionEvent when the
     * user attempts to close the window. The window may be
     * closed by clicking on "X", choosing Close from the
     * window menu, or pressing the Escape key. The default
     * command is "CloseRequested", which is just what a Close
     * choice button would probably have as a command.
     *
     * @param command        used when window closed
     */
    public void setCloseWindowCommand(String command) {
        msCloseWindowCommand = command;
    }

    public void setLocation(int x, int y) {
        this.mX = x;
        this.mY = y;
    }

    /**
     * A convenience method that assumes the command is the
     * same as the label and no callback is desired.
     *
     * @param label                Button's caption and unique name
     */
    public void addButton(String label) {
        addButton(label, "");
    }

    /**
     * Add a button to the dialog
     *
     * @param label                Button's caption (Must be unique on dialog)
     * @param callbackMethod    Method to call when button pushed (may be "") if
     *                            no callback is desired.
     */
    public void addButton(String label, String callbackMethod) {
        Button button = new Button(label);
        button.setActionCommand(label);
        button.addActionListener(this);
        button.addKeyListener(this);
        button.addFocusListener(this);

        //resize layout to accomodate a new button
        GridLayout gridLayout = (GridLayout) moButtonPanel.getLayout();
        gridLayout.setColumns(gridLayout.getColumns() + 1);
        moButtonPanel.setLayout(gridLayout);

        moButtonPanel.add(button);

        ButtonAction buttonA = new ButtonAction(label, callbackMethod);
        mvButtons.addElement(buttonA);
    }

    /**
     * Set the default button.  Only one button can be the default, so if 
     * you call this more than once, only the last one called will be the default).
     *
     * @param label                Button's caption 
     */
    public void setDefaultButton(String label) {
        msDefaultButton = label;
    }

    /**
     * A convenience method that assumes the key is the
     * same as the Question and uses defaults.
     */
    public void addQuestion(String question) {
        addQuestion(question, question, "", true, false, null);
    }

    /**
     * A convenience method that uses defaults.
     */
    public void addQuestion(String question, String key) {
        addQuestion(question, key, "", true, false, null );
    }

    /**
     * A convenience method that uses defaults.
     */
    public void addQuestion(String question, String key, String defaultAnswer) {
        addQuestion(question, key, defaultAnswer, true, false, null);
    }

    public void addQuestion(String question, String key, String defaultAnswer, boolean enabled, boolean password ) {
        addQuestion(question, key, defaultAnswer, enabled, password, null);
    }
    /**
     * Add a question (a lable with the text of the question and a
     * textfield below used for the answer) to the dialog.
     *
     * @param question            what to display in the label above the textfield
     * @param key                a unique identifier that is used to construct the XML and used
     *                            in getAnswer()
     * @param defaultAnswer        value to put into the textfield ("" is an acceptable input)
     * @param enabled            if true the textfield will allow edits 
     * @param password            if true the echo char will be set to '*'
     */
    public void addQuestion(String question, String key, String defaultAnswer, boolean enabled, boolean password, String restricted) {
        Label lbl = new Label(" " + question);
        Label lblAnswer = new Label(" " + defaultAnswer);
        lbl.setBackground(moColor);
        TextField txt = new TextField(defaultAnswer, 50);

        if (enabled) {
            Question Q = new Question(question, defaultAnswer, key, "textfield" + miQNbr++, false, txt);
            mvQuestions.addElement(Q);

            txt.setBackground(Color.white);
            txt.addKeyListener(this);

            // Used to find restricted characters in key listener.
            txt.setName(restricted);

            if (password) {
                txt.setEchoChar('*');
            }
        } else {
            //Use a label to make it clear that the value is readonly
            //Font oBoldFont = new Font(moDialog.getFont().getName(), Font.BOLD, moDialog.getFont().getSize());

            lblAnswer.setBackground(moColor); //Color.white);
            //lblAnswer.setFont(oBoldFont);
            addHidden(key, defaultAnswer);
        }

        //resize layout to accomodate a two new rows
        GridLayout gridLayout = (GridLayout) moInputPanel.getLayout();
        if (gridLayout.getRows() == 1)
            gridLayout.setRows(gridLayout.getRows() + 1);
        else
            gridLayout.setRows(gridLayout.getRows() + 2);

        moInputPanel.setLayout(gridLayout);

        moInputPanel.add(lbl);
        if (enabled)
            moInputPanel.add(txt);
        else
            moInputPanel.add(lblAnswer);
    }

    /**
     * Add a caption (without a TextField below) to the InputBox that will appear in
     * the order called.  Useful if you want to have more than one line of text for
     * a question.
     *
     * @param caption            what to display in the label 
     */
    public void addCaption(String caption) {
        addCaption(caption, false);
    }

    /**
     * Add a caption (without a TextField below) to the InputBox that will appear in
     * the order called.  Useful if you want to have more than one line of text for
     * a question.
     *
     * @param caption            what to display in the label 
     * @param bold              display in bold? 
     */
    public void addCaption(String caption, boolean bold) {
        Label lbl = new Label(" " + caption);
        if( bold ) {
            lbl.setFont(boldFont);
        }
        lbl.setBackground(moColor);

        //resize layout to accomodate a new row
        GridLayout gridLayout = (GridLayout) moInputPanel.getLayout();
        gridLayout.setRows(gridLayout.getRows() + 1);
        moInputPanel.setLayout(gridLayout);

        moInputPanel.add(lbl);
    }

    /**
     * Add a List control with a caption to the botton of an InputBox.  This should be the 
     * last thing called when constructing an Inputbox so that the caption will appear
     * in the correct place.
     *
     * @param caption            what to display in the label above the list control
     * @param key                a unique identifier that is used to construct the XML and used
     *                            in getAnswer()
     * @param XMLItems            XML representing the contents of the list control and 
     *                            what value should be returned when the item is selected
                                 Example: "<Item><Display>Apple</Display><Value>A1</Value></Item><Item><Display>Banana</Display><Value>B1</Value></Item>"
     * @param rows                The number of visible rows in the list control 
     */
    public void addList(String caption, String key, String XMLItems, int rows) {
        addList(caption,key,XMLItems,rows,false);
    }
    public void addList(String caption, String key, String XMLItems,
                        int rows, boolean multiple) {
        /* =================================================
            XML FORMAT...
                <Item>
                    <Display></Display>
                    <Value></Value>
                </Item>
                <Item>
                    <Display></Display>
                    <Value></Value>
                </Item>
                ...
           ================================================= */

        //Add caption to the list control.  This won't display properly if addQuestion is called after this
        addCaption(caption);

        List lst = new List(rows);
        if( multiple )
            lst.setMultipleMode(true);

        MiniXMLParser oParseRow = new MiniXMLParser();
        MiniXMLParser oParseVal = new MiniXMLParser();
        String sRow;
        String sValue;
        String sDisplay;

        //Parse the XML and add the values to the list box and a
        //vector that contains the values that should be returned
        sRow = oParseRow.parseTag(XMLItems, "Item");
        while (sRow.length() > 0) {
            sDisplay = oParseVal.parseTag(sRow, "Display");
            lst.addItem(sDisplay);
            sValue = oParseVal.parseTag(sRow, "Value");
            if (sValue.length() > 0)
                mvListValues.addElement(sValue);
            else
                //Use what is displayed for the value since none supplied
                mvListValues.addElement(sDisplay);
            sRow = oParseRow.parseNextTag(XMLItems, "Item");
        }

        moListPanel = new Panel();
        moListPanel.setLayout(new GridLayout(1,1));
        moListPanel.setBackground(moColor);
        moListPanel.add(lst);

        Question Q = new Question(caption, "", key, "list", false, lst);
        mvQuestions.addElement(Q);

        //Select the first item by default
        if (lst.getItemCount() > 1)
            lst.select(0);

    }

    /**
     * Add a hidden value to dialog.  This value isn't displayed but is returned 
     * in the XML string and in the callback method and can be retrieved using
     * the getAnswer() method.  This is used to help retain state information
     *
     * @param key                a unique identifier that is used to construct the XML and used
     *                            in getAnswer()
     * @param value                value of this key
     */
    public void addHidden(String key, String value) {
        Question Q = new Question("", value, key, "Hidden", true, null);
        mvQuestions.addElement(Q);
    }

    /**
     * Display inputbox dialog
     */
    public void inputbox() {
        displayDialog("", true);
    }

    /**
     * Display msgbox dialog with the specified question/statment
     *
     * @param message    Message to display 
     */
    public void msgbox(String message) {
        displayDialog(message, false);
    }

    /**
     * Same as msgbox(String message) except adds an "OK" button.
     * (no callback method will be called)
     */
    public void msgboxOK(String message) {
        addButton("OK", "");
        msDefaultButton = "OK";
        msgbox(message);
    }

    /**
     * Same as msgbox(String message) except adds "Yes" and "No"
     * buttons.  (no callback method will be called)
     */
    public void msgboxYesNo(String message) {
        addButton("Yes", "");
        addButton("No", "");
        msgbox(message);
    }

    /**
     * retrieve the contents of any of the textfields or hidden values
     * by suppling the unique key that represents the answer 
     * (the same key supplied in addQuestion() or addHidden() )
     */
    public String getAnswer(String key) {
        Question Q;
        Enumeration enum = mvQuestions.elements();

        while (enum.hasMoreElements()) {
            Q = (Question)enum.nextElement();
            if (Q.getKey().equals(key)) {
                return Q.getAnswer();
            }
        }
        return "";
    }

    //---------- Private Methods -----------------------------

    /**
     * Display the dialog in either inputbox or msgbox modes
     *
     * @param message        Message to display (msgbox mode)
     * @param inputbox        if true, will display an inputBox dialog
     *                        otherwise will display a msgbox dialog    
     */
    private void displayDialog(String message, boolean inputbox) {
        if (moFrame == null) {
            moFrame = new Frame();
            mbframeNotProvided = true;
        } else {
            mbframeNotProvided = false;
        }
        moDialog = new Dialog(moFrame, true); // Modal
        moDialog.addWindowListener(this);
        moDialog.addKeyListener(this);
        moDialog.setTitle(msTitle);
        moDialog.setLayout(new BorderLayout(5, 5));
        moDialog.setBackground(moColor);
        moDialog.setFont(moFont);

        //Add extra blank label so it's not so squished
        Label lblSpacer = new Label();
        lblSpacer.setBackground(moColor);
        moButtonPanel.add(lblSpacer);

        //Add extra row of blanks so it's not so squished
        GridLayout buttonPanelLayout = (GridLayout)moButtonPanel.getLayout();
        int i = buttonPanelLayout.getColumns();
        while( i-- > 0 ) {
            lblSpacer = new Label();
            lblSpacer.setBackground(moColor);
            moButtonPanel.add(lblSpacer);
        }

        if (inputbox) {
            if (moListPanel != null) {
                moDialog.add("North", moInputPanel);
                moDialog.add("Center", moListPanel);
            } else {
                addCaption(""); //Add extra blank label so it's not so squished
                moDialog.add("Center", moInputPanel);
            }
        } else {
            Panel messagePanel = createMultiLinePanel(" " + message);
            messagePanel.setBackground(moColor);
            moDialog.add("Center", messagePanel);
        }

        moDialog.add("South", moButtonPanel);

        moDialog.pack();
        moDialog.setBackground(moColor);
        enforceMinimumSize(moDialog, 200, 100);
        centerWindow(moDialog);

        // Start a new thread to show the dialog
        Thread thread = new Thread(this);
        thread.start();
    }

    /**
     * Close the dialog and call the appropriate
     * callback method for the clicked button
     *
      * @param clickedButton            Caption of the clicked button
     */
    private void closeDialogAndCallBack(String clickedButton) {
        //close dialog
        moDialog.setVisible(false);
        moDialog.dispose();
        if (mbframeNotProvided)
            moFrame.dispose();

        if (moCallback != null) {
            //call the method for the clicked
            //button and pass an XML representation of the
            //users input
            callbackMethod(clickedButton);
        }
    }



    /**
     * Calls the method in the Listener that was specified
     * in the addButton method in the callbackMethod parameter
     *
     * @param command    the name of the button that was clicked    
     */
    private void callbackMethod(String buttonLabel) {
        ButtonAction buttonA;
        Enumeration enum = mvButtons.elements();

        while (enum.hasMoreElements()) {
            buttonA = (ButtonAction)enum.nextElement();
            if (buttonA.getLabel().equals(buttonLabel)) {
                try {
                    Class[] argumentTypes = { String.class };
                    Object[] argumentValue = {constructAnswerXML(buttonLabel)};

                    Method method = moCallback.getClass().getMethod(buttonA.getCallbackMethod(), argumentTypes);
                    Object result = method.invoke(moCallback, argumentValue);
                } catch (Exception e) {
                    //An exception will be raised if the method name is empty or doesn't match
                    //ignore error
                }
                return;
            }
        }
    }

    /**
     * Constructs an XML string that contains the questions (keys)
     * and the answers supplied along with the hidden values and
     * the button that was clicked.  
     *
     * @param button    the name of the button that was clicked    
     */
    private String constructAnswerXML(String buttonLabel) {
        StringCleanup oStringCleanup = new StringCleanup();
        String sXML = "<Answer>";
        Question Q;
        Enumeration enum = mvQuestions.elements();

        while (enum.hasMoreElements()) {
            Q = (Question)enum.nextElement();

            //Put each questions key and answer into its own tag
            sXML += "<" + Q.getKey() + ">"
                    /* make sure the answers are XML friendly */
                    +  oStringCleanup.Cleanup(Q.getAnswer(), false)
                    + "</" + Q.getKey() + ">";

        }
        //Add which button was clicked
        sXML += "<Button>" + buttonLabel + "</Button>";

        sXML += "</Answer>";

        return sXML;
    }

    // Used to create a word wrapping message
    private Panel createMultiLinePanel(String message) {
        Panel mainPanel = new Panel();
        mainPanel.setBackground(moColor);
        GridBagLayout gbLayout = new GridBagLayout();
        mainPanel.setLayout(gbLayout);
        addMultilineString(message, mainPanel);
        return mainPanel;
    }

    // There are a variaty of ways to do this....
    private void addMultilineString(String message, Container container) {

        GridBagConstraints constraints = getDefaultConstraints();
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        // Insets() args are top, left, bottom, right
        constraints.insets   = new Insets(0,0,0,0);
        GridBagLayout gbLayout = (GridBagLayout)container.getLayout();

        //pad with one line before and one line after msg
        Label labelBefore = new Label();
        labelBefore.setBackground(moColor);
        gbLayout.setConstraints(labelBefore, constraints);
        container.add(labelBefore);
        constraints.anchor = GridBagConstraints.CENTER;
        while (message.length() > 0) {
            int newLineIndex = message.indexOf('\n');
            String line;
            if (newLineIndex >= 0) {
                line = message.substring(0, newLineIndex);
                message = message.substring(newLineIndex + 1);
            } else {
                line = message;
                message = "";
            }
            Label label = new Label(line);
            label.setBackground(moColor);
            gbLayout.setConstraints(label, constraints);
            container.add(label);
        }

        //pad with one line before and one line after msg
        //so that it's easier to read
        //(2nd label added in displayDialog)
        Label labelAfter = new Label();
        labelAfter.setBackground(moColor);
        gbLayout.setConstraints(labelAfter, constraints);
        container.add(labelAfter);
    }

    //Set once use multiple times
    private GridBagConstraints getDefaultConstraints() {
        GridBagConstraints constraints = new GridBagConstraints();
        constraints.weightx  = 1.0;
        constraints.weighty  = 1.0;
        constraints.gridheight = 1; // One row high
        // Insets() args are top, left, bottom, right
        constraints.insets   = new Insets(4,4,4,4);
        // fill of NONE means do not change size
        constraints.fill = GridBagConstraints.NONE;
        // WEST means align left
        constraints.anchor = GridBagConstraints.WEST;

        return constraints;
    }

    //Center Dialog Window on screen
    private void centerWindow(Window win) {
        Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
        // If larger than screen, reduce window width or height
        if (screenDim.width < win.getSize().width) {
            win.setSize(screenDim.width, win.getSize().height);
        }
        if (screenDim.height < win.getSize().height) {
            win.setSize(win.getSize().width, screenDim.height);
        }
        // Center Frame, Dialogue or Window on screen if no coordinates specified
        if (mX < 0 ) {
            win.setLocation(mX, mY);
        } else {
            int x = (screenDim.width - win.getSize().width) / 2;
            int y = (screenDim.height - win.getSize().height) / 2;
            win.setLocation(x, y);
        }
    }

    //Enforce minimum size
    private void enforceMinimumSize(Component comp,
                                    int minWidth, int minHeight) {
        if (comp.getSize().width < minWidth) {
            comp.setSize(minWidth, comp.getSize().height);
        }
        if (comp.getSize().height < minHeight) {
            comp.setSize(comp.getSize().width, minHeight);
        }
    }

    //===========================================================================
    //                   I N T E R N A L      C L A S S E S
    //===========================================================================

    /**
     * Contains info that relates to a specific button.  This class is used to 
     * create objects that can be stored inside of a vector.
     *
     * @author Jonathan Paulson
     */
    public class ButtonAction {
        private String        msLabel;
        private String        msCallbackMethod;

        public ButtonAction(String label, String callbackMethod) {
            this.msLabel = label;
            this.msCallbackMethod = callbackMethod;
        }

        public String getLabel() {
            return msLabel;
        }

        public String getCallbackMethod() {
            return msCallbackMethod;
        }
    }

    /**
     * Contains info that relates to a specific question.  This class is used to 
     * create objects that can be stored inside of a vector.
     *
     * @author Jonathan Paulson
     */
    public class Question {
        String        msQuestion    = "";
        String        msAnswer    = "";
        String        msKey        = "";
        String        msName        = "";
        boolean        mbHidden    = false;
        Object      theObject   = null;

        public Question(String question, String answer, String key, String name, boolean hidden, Object o) {
            this.msQuestion = question;
            this.msAnswer = answer;
            this.msKey = key;
            this.msName = name;
            this.theObject = o;
        }

        public String getQuestion() {
            return msQuestion;
        }

        public String getAnswer() {
            if( theObject instanceof TextField ) {
                msAnswer = ((TextField)theObject).getText();
            } else if( theObject instanceof List ) {
                String sValue = "";
                int[] iIndexes = ((List)theObject).getSelectedIndexes();
                for( int i = 0; i < iIndexes.length; i++ ) {
                    int j = iIndexes[i];
                    if( i > 0 )
                        sValue += ":";
                    sValue += (String) mvListValues.elementAt(j);
                }
                msAnswer = sValue;
            }
            return msAnswer;
        }

        public String getKey() {
            return msKey;
        }

        public String getName() {
            return msName;
        }

        public boolean hidden() {
            return mbHidden;
        }

        public void setAnswer(String answer) {
            msAnswer = answer;
        }
    }
}
